@@ -25,7 +25,7 @@ Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 module Text.TeXMath.Readers.TeX (readTeX)
 where
 
-import Data.List (intercalate, intersperse, find)
+import Data.List (intercalate, intersperse, find, foldl')
 import Control.Monad
 import Data.Char (isDigit, isAscii, isLetter)
 import qualified Data.Map as M
@@ -45,6 +45,7 @@ import Text.TeXMath.Shared (getSpaceChars)
 import Data.Generics (everywhere, mkT)
 import Text.TeXMath.Readers.TeX.Commands ( styleOps, textOps, enclosures,
                                            operators, symbols, siUnitMap )
+import Data.Text.Read (decimal)
 
 type TP = Parser
 
@@ -67,40 +68,57 @@ expr1 = choice
 readTeX :: Text -> Either Text [Exp]
 readTeX inp =
   let (ms, rest) = parseMacroDefinitions inp in
-  either (Left . showParseError inp) (Right . everywhere (mkT fixBins))
+  either (Left . showParseError inp) (Right . fixBinList)
     $ parse formula "formula" $ applyMacros ms rest
-
--- | Convert Bin after nothing, Open, Pun, or Op to Op (#176).
-fixBins :: Exp -> Exp
-fixBins e =
-  case e of
-    EGrouped es -> EGrouped (fixBinList True es)
-    EDelimited op cl des -> EDelimited op cl (fixBinListDel True des)
-    _ -> e
  where
-  fixBinList atBeginning xs =
-    case xs of
-      ESymbol Bin t : rest
-        | atBeginning
-        -> ESymbol Op t : fixBinList False rest
-      ESymbol Bin t : rest@(ESymbol ty _ : _)
-        | ty == Open || ty == Pun || ty == Op
-        -> ESymbol Op t : fixBinList False rest
-      x:rest -> x : fixBinList False rest
-      [] -> []
+  -- | Convert Bin symbol type in certain contexts (#176, #234).
+  fixBinList :: [Exp] -> [Exp]
+  fixBinList =
+    reverse . foldl' goExp [] . everywhere (mkT fixBins)
 
-  fixBinListDel atBeginning xs =
-    case xs of
-      Left x : rest
-        ->  Left x : fixBinListDel True rest
-      Right (ESymbol Bin t) : rest
-        | atBeginning
-        -> Right (ESymbol Op t) : fixBinListDel False rest
-      Right (ESymbol Bin t) : rest@(Right (ESymbol ty _):_)
-        | ty == Open || ty == Pun || ty == Op
-        -> Right (ESymbol Op t) : fixBinListDel False rest
-      x:rest -> x : fixBinListDel False rest
-      [] -> []
+  -- TeXBook:
+  -- 5. If the current item is a Bin atom, and if this was the first
+  -- atom in the list, or if the most recent previous atom was Bin, Op,
+  -- Rel, Open, or Punct, change the current Bin to Ord and continue with
+  -- Rule 14. Otherwise continue with Rule 17.
+  -- 6. If the current item is a Rel or Close or Punct atom, and if
+  -- the most recent previous atom was Bin, change that previous Bin
+  -- to Ord. Continue with Rule 17.
+
+  fixBins :: Exp -> Exp
+  fixBins e =
+    case e of
+      EGrouped es
+        -> EGrouped (reverse $ foldl' goExp [] es)
+      EDelimited op cl des
+        -> EDelimited op cl (reverse $ foldl' goInDel [] des)
+      EArray als alines
+        -> EArray als (map (map (reverse . foldl' goExp [])) alines)
+      EStyled tt es
+        -> EStyled tt (reverse $ foldl' goExp [] es)
+      _ -> e
+
+  goExp :: [Exp] -> Exp -> [Exp]
+  goExp [] (ESymbol Bin t) = [ESymbol Ord t]
+  goExp  accum@(ESymbol ty _ : _) (ESymbol Bin t)
+    | ty `elem` [Bin, Op, Rel, Open, Pun]
+      = ESymbol Ord t : accum
+  goExp (ESymbol Bin t' : rest) (ESymbol ty t)
+    | ty `elem` [Rel, Close, Pun]
+      = ESymbol ty t : ESymbol Ord t' : rest
+  goExp xs x = x : xs
+
+  goInDel :: [InEDelimited] -> InEDelimited -> [InEDelimited]
+  goInDel [] (Right (ESymbol Bin t)) = [Right (ESymbol Ord t)]
+  goInDel accum@(Left _ : _) (Right (ESymbol Bin t))
+    = Right (ESymbol Ord t) : accum
+  goInDel accum@(Right (ESymbol ty _) : _) (Right (ESymbol Bin t))
+    | ty `elem` [Bin, Op, Rel, Open, Pun]
+      = Right (ESymbol Ord t) : accum
+  goInDel (Right (ESymbol Bin t') : rest) (Right (ESymbol ty t))
+    | ty `elem` [Rel, Close, Pun]
+      = Right (ESymbol ty t) : Right (ESymbol Ord t') : rest
+  goInDel xs x = x : xs
 
 showParseError :: Text -> ParseError -> Text
 showParseError inp pe =
@@ -142,6 +160,7 @@ ignorable = skipMany $
     <|> tag
     <|> () <$ ctrlseq "nonumber"
     <|> (skipMany1 space <?> "whitespace")
+    <|> (() <$ ctrlseq "allowbreak")
 
 comment :: TP ()
 comment = char '%' *> skipMany (noneOf "\n") *> optional newline
@@ -183,6 +202,7 @@ command = try $ do
   choice
     [ text c
     , styled c
+    , colored c
     , root c
     , xspace c
     , mathop c
@@ -388,19 +408,17 @@ delimited = do
 
 delimitedImplicit :: TP Exp
 delimitedImplicit = try $ do
-  openc <- lexeme $ oneOf "()[]|"
-  closec <- case openc of
-                 '(' -> return ')'
-                 '[' -> return ']'
-                 '|' -> return '|'
-                 _   -> mzero
-  let closer = lexeme $ char closec
+  (opent, closer) <-
+           (("(", symbol ")") <$ symbol "(")
+       <|> (("[", symbol "]") <$ symbol "[")
+       <|> (("|", symbol "|") <$ symbol "|")
+       <|> (("\x2016", "\x2016" <$ ctrlseq "rVert") <$ ctrlseq "lVert")
   contents <- concat <$>
               many (try $ ((:[]) . Left  <$> middle)
                       <|> (map Right . unGrouped <$>
                              many1Exp (notFollowedBy closer *> expr)))
-  _ <- closer
-  return $ EDelimited (T.singleton openc) (T.singleton closec) contents
+  closet <- T.pack <$> closer
+  return $ EDelimited opent closet contents
 
 scaled :: Text -> TP Exp
 scaled cmd = do
@@ -424,23 +442,35 @@ endLineAMS = lexeme $ try $ do
   return '\n'
 
 arrayLine :: TP ArrayLine
-arrayLine = notFollowedBy (ctrlseq "end" >> return '\n') >>
-  sepBy1 (unGrouped <$>
-    manyExp (try (ignorable' *> notFollowedBy endLine) *>
-               expr <*
-               ignorable')) (symbol "&")
-  where ignorable' = ignorable >>
-                     optional (try (ctrlseq "hline" >> ignorable'))
+arrayLine =
+  sepBy1
+    (ignorable' *>
+      (unGrouped <$>
+        manyExp (try (notFollowedBy (('\n' <$ ctrlseq "end") <|> endLine)) *>
+                   expr <* ignorable'))) (symbol "&")
+  where ignorable' = ignorable >> optional (try (ctrlseq "hline" >> ignorable'))
   -- we don't represent the line, but it shouldn't crash parsing
 
 arrayAlignments :: TP [Alignment]
-arrayAlignments = try $ do
-  as <- braces (many (letter <|> char '|'))
-  let letterToAlignment 'l' = AlignLeft
-      letterToAlignment 'c' = AlignCenter
-      letterToAlignment 'r' = AlignRight
-      letterToAlignment _   = AlignCenter
-  return $ map letterToAlignment $ filter (/= '|') as
+arrayAlignments = mconcat <$>
+  braces (many (
+                ((:[]) . letterToAlignment <$> letter)
+            <|> ([] <$ char '|')
+            <|> ([] <$ oneOf " \t")
+            <|> ([] <$ ((char '@' <|> char '!') <* inbraces))
+            <|> (do char '*'
+                    num <- T.pack <$> braces (many1 digit)
+                    cols <- arrayAlignments
+                    case decimal num of
+                      Left msg -> fail msg
+                      Right (n :: Int, _)
+                               -> return $ mconcat $ replicate n cols)
+               ))
+ where
+   letterToAlignment 'l' = AlignLeft
+   letterToAlignment 'c' = AlignCenter
+   letterToAlignment 'r' = AlignRight
+   letterToAlignment _   = AlignCenter
 
 environment :: Text -> TP Exp
 environment "\\begin" = do
@@ -477,6 +507,7 @@ environments = M.fromList
   , ("Vmatrix", matrixWith "\x2225" "\x2225")
   , ("split", align)
   , ("multline", gather)
+  , ("multlined", gather)
   , ("gather", gather)
   , ("gathered", gather)
   , ("equation", equation)
@@ -488,7 +519,7 @@ alignsFromRows defaultAlignment (r:_) = replicate (length r) defaultAlignment
 
 matrixWith :: Text -> Text -> TP Exp
 matrixWith opendelim closedelim = do
-  lines' <- sepEndBy1 arrayLine endLineAMS
+  lines' <- sepEndBy arrayLine endLineAMS
   let aligns = alignsFromRows AlignCenter lines'
   return $ if T.null opendelim && T.null closedelim
               then EArray aligns lines'
@@ -543,14 +574,13 @@ variable = do
 isConvertible :: Exp -> Bool
 isConvertible (EMathOperator x) = x `elem` convertibleOps
   where convertibleOps = [ "lim","liminf","limsup","inf","sup"
-                         , "min","max","Pr","det","gcd"
-                         ]
-isConvertible (ESymbol Rel _) = True
-isConvertible (ESymbol Bin _) = True
+                         , "min","max","Pr","det","gcd" ]
 isConvertible (ESymbol Op x) = x `elem` convertibleSyms
-  where convertibleSyms = ["\x2211","\x220F","\x22C2",
-           "\x22C3","\x22C0","\x22C1","\x2A05","\x2A06",
-           "\x2210","\x2A01","\x2A02","\x2A00","\x2A04"]
+  where convertibleSyms =
+         ["\x2211", "\x220F", "\x2210", -- \sum \prod \amalg
+          "\x22C0", "\x22C1", -- bigwedge \bigvee
+          "\x22C2", "\x22C3", -- \bigcap \bigcup
+          "\x2A05", "\x2A06"] -- \bigsqcap \bisqcup
 isConvertible _ = False
 
 -- check if sub/superscripts should always be under and over the expression
@@ -648,6 +678,13 @@ styled c = do
                        _           -> f [x]
        Nothing  -> mzero
 
+colored :: Text -> TP Exp
+colored "\\color" = do
+  _ <- inbraces -- skip the color
+  -- in the future we might add color to the types or to the styles
+  texSymbol <|> inbraces <|> texChar
+colored _ = mzero
+
 -- note: sqrt can be unary, \sqrt{2}, or binary, \sqrt[3]{2}
 root :: Text -> TP Exp
 root c = do
@@ -655,8 +692,8 @@ root c = do
   (ERoot <$> inbrackets <*> texToken) <|> (ESqrt <$> texToken)
 
 xspace :: Text -> TP Exp
-xspace "\\mspace" = do
-  ctrlseq "mspace"
+xspace "\\enspace" = return $ ESpace (1/2)
+xspace "\\mspace" =
   braces $ do
     len <- many1 digit
     lexeme $ string "mu"
